package outils;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

// This algorithm runs the coalescence under a model of structured archipelago: it builds (backward in time) 
// the genealogy of a sample of the individuals living within the archipelago at present time, under neutrality hypotheses, 
// and following the migration of their ancestral lineages among islands back in time, until reaching the immigration events from the mainland.
// Input files and output files are read/produces using "ReadTable.java"; explanations about inputs and outputs are given in this java file.
// One "coaHist_struct_test_SIM.txt" (history of coalescence) file is produced for each simulation replicate SIM (if all results were put in 1 file, the latter would be too big to be used in the second step). 
// Fix the number of simulation replicates, nbSim, as you wish.
// The word "patch" used below refers to individual islands.

// In this study example, we consider simplified input files (few islands, few individuals, simple theoretical matrix of connectivity between landmasses), 
// named "name_test.txt". Once you understand the code, make your own input files.
// We left (inactivated, commented) commands System.out.println(): they enable to print events on screen, and you may want to re-activate them to better understand how the code works.
// If you want to consider multiple independent structured archipelago, (e.g. at increasing distance to the mainland), you need to reproduce the process with another matrix of connectivity, and thus another set of output files "coaHist_struct_..._SIM.txt".

public class Coalescent {

// ATTRIBUTES:

	int J;                     			// Number of individuals in the archipelago
	int N;                      		// Total number of islands
	int[] CCPat;                		// Carrying capacity of the islands
	double[] freq;						// Sampling frequencies in each island
	double[][] Mcon;					// Matrix of connectivity between islands
	int[] distri;						// Mainland species abundance distribution
	static int time;					// time step						
	static Individual[] sample;			// Spatialized sample of individuals followed by the coalescence
	static Individual target;			// Individual randomly chosen at each time step, target individual

	static int event;					// Informations to write the output file
	static int from;					
	static int motherNum;
	static int position;

	LinkedList<Individual> activ;       // List of the individuals in the coalescent ("active" individuals; list updated at each time step; its size decreases as time goes backward)
	static int nbActiv;                 // Number of individuals in this list 
	int[] actPat;                       // Number of individuals in the coalescent which are located in island i

	static double[] transition;			// Vector of transition probabilities, which allows simulating the markovian dynamic of the coalescent. 
										// Knowing which individual is just born, these probabilities are
											// 0 -> probability of local coalescence: the mother of the born individual is another traced individual, from the same island
											// 1 -> probability of local recruitment: the mother of the born individual is an individual not traced (not sampled), from the same island
											// 2 -> probability of migration from the mainland: the mother of the born individual is an individual from the mainland
											// i : [3,(N+1)] ->  probability of external recruitment: the mother of the born individual is an individual not traced (not sampled), from another island
											// i : [(N+2),(2N)] -> probability of external coalescence: the mother of the born individual is another traced individual, from another island

	static int nbSim = 10;						// Number of simulation replicates: TO BE MODIFIED
	
// CONSTRUCTOR:

	public Coalescent() throws Exception{
		
		Mcon = ReadTable.readConnectivity();		
		distri = ReadTable.readDistri();		
		CCPat = ReadTable.readK();
		freq = ReadTable.readSamplingFreq();
		
		N = Mcon.length-1;											// N = number of islands	
		int[] nbSampl = new int[N]; 
		int tailleSampl = 0;		
		int tailleMetaCom = 0;
		for(int i = 0 ; i<N ; i++){
			nbSampl[i] = (int)Math.floor(CCPat[i]*freq[i]);			// Sampled individuals
			tailleSampl = tailleSampl+nbSampl[i];
			tailleMetaCom = tailleMetaCom+CCPat[i];
			if(nbSampl[i]>CCPat[i]) throw(new Exception("Over sampling of island "+i));
		}
		J = tailleMetaCom;				

		nbActiv = tailleSampl;
		actPat = new int[N];
		for(int i = 0 ; i<N ; i++){
			actPat[i]=nbSampl[i];
		}
		activ = new LinkedList<Individual>();
		Individual[] samp = new Individual[tailleSampl];
		int curseur = 0;
		for(int i = 0 ; i<N ; i++){
			for(int j = 0 ; j<nbSampl[i] ; j++){
				Individual ind = new Individual(i,curseur);
				Lineage lign = new Lineage(ind);
				activ.add(ind);
				samp[curseur]=ind;
				curseur++;
			}
		}
		sample = samp;		
		target = sample[0];
		motherNum = 0;
		event = 0;
		from = 0;
		position = -1;
		time = 0;
		transition = new double[2*N+2];
	}

// READERS:

	public int nbActifs(){
		return(nbActiv);
	}

// METHODS:

	public void cycle() throws Exception{
		
		time++;
		position=-1;
		if (nbActiv<=0) throw(new Exception("Coalescence already finished"));
		
		else if(nbActiv==1){			// If the coalescent contains only one individual, we look for the mainland ancestor to this individual.
			target = activ.get(0);			// the target individual is the last individual within the list of active individuals
			activ.remove(0);				// list "activ" updated
			nbActiv--;					
			int iPC = target.getLoc();		// iPC = index of the island of the target individual
			actPat[iPC]--;              	// list "actPat" updated
			event = 3;						// Migration from the mainland
			from = target.getNum();
			
			Individual mother = new Individual(target.getLoc(),target.getNum());	// Choosing the mainland ancestor
			target.getLineage().descendFrom(mother);
			mother.getLineage().newMainlandSpecies(mother, distri);
			
			mother.getLineage().adjustTime(time); 
			//System.out.println("Mainland species: ");
			//System.out.println(mother.getSpc());
		}

		else{

			// STEP 1 : Choice of the target individual

			double r = Math.random();
			int ind_cible = (int)Math.floor(r*(double)nbActiv); 	// Integer within [0,nbActiv-1]: index of the target individual
			target = activ.get(ind_cible);

			// STEP 2 : Calculation of the transition vector, knowing which individual is just born
			
			int iPC = target.getLoc();								// iPC = index of the island of the target individual
			int KPC = CCPat[iPC];									// Carrying capacity of the island of the target individual
			
			double m = Mcon[iPC][iPC];								// Probability of coming from the same island		
			transition[0] = m*(actPat[iPC]-1)/((double) (KPC-1));		// Local coalescence
			transition[1] = m*(KPC-actPat[iPC])/((double) (KPC-1));		// Local recruitment
			transition[2] = Mcon[iPC][N];								// Migration from the mainland
			int countPatch=0;
			for (int numIsland = 0; numIsland<N; numIsland++){ 			//For all islands but iPC
				if (numIsland != iPC) {
					transition[3+countPatch]=Mcon[iPC][numIsland]*(CCPat[numIsland]-actPat[numIsland])/(double)CCPat[numIsland];		// Recruitment with mother in patch (numIsland)
					transition[2+N+countPatch]=Mcon[iPC][numIsland]*(actPat[numIsland]/(double)CCPat[numIsland]);						// Coalescence with mother in patch (numIsland)
					countPatch++;
				}
			}
			//System.out.println("Index of the target individual:");
			//System.out.println(target.getNum());
			
/*			System.out.println("Transition probabilities:");
			System.out.println(transition[0]);	// Local coalescence
			System.out.println(transition[1]);	// Local recruitment
			System.out.println(transition[2]);	// Migration from the mainland
			System.out.println(transition[3]);	// External recruitment
			System.out.println(transition[4]);  // External recruitment...
			System.out.println(transition[5]);  // External coalescence
			System.out.println(transition[6]);	// External coalescence... */
			
			// STEP 3 : Choice of the transition
			
			r = Math.random();
			int saut = 0;
			int count = 0;
			double somme_cum = transition[0];
			while(somme_cum<r){
				saut++;
				somme_cum = somme_cum+transition[saut]; 
			}
			
			// STEP 4 : Update of the system according to the transition
			
			if(saut == 0){
				//System.out.println("Local coalescence");
				r = Math.random();
				int iDPP = (int)Math.floor(r*(actPat[iPC]-1));  // Index of the active mother, within the non-target active individuals of the island iPC (integer within [0,actPat[iPC]-1])
				activ.remove(ind_cible);
				nbActiv--;
				actPat[iPC]--;
				Iterator<Individual> it = activ.iterator();
				int compte = -1;
				Individual mother = target; 					// Initial value without any importance
				boolean trouve = false;
				while(!trouve){
					if(!it.hasNext()) throw(new Exception("Wrong assessment of the number of active individuals inside island "+iPC));
					else mother = it.next();
					if(mother.getLoc() == iPC){
						compte++;
					}
					trouve = (compte == iDPP);
				}
				//System.out.println("ancestor.getNum:");
				//System.out.println(mother.getNum());
				event = 2;
				motherNum = mother.getNum();
				from = target.getNum();
				position = target.getLoc();
				mother.getLineage().mother(target.getLineage().getSampl());
				for (int i1 = 0; i1<mother.getLineage().getSampl().size();i1++){
					if (mother.getLineage().getSampl().get(i1).getNum() != target.getNum()){
						sample[target.getNum()].setTopInd(-1);
					}
				}
				
			}
			else if(saut == 1){				
				event = 0;
				from = 0;
				//System.out.println("Local recruitment");		// The mother has the same characteristics as the target individual, so nothing changes in the macro data of the coalescent (the 2 following lines are thus useless)
				Individual mother = new Individual(target.getLoc(),target.getNum());
				target.getLineage().descendFrom(mother);				
			}
			
			else if (saut == 2){
				event = 3;
				from = target.getNum();
				//System.out.println("Migration from the mainland");	
				actPat[iPC]--;
				activ.remove(ind_cible);
				nbActiv--;	
				Individual mother = new Individual(target.getLoc(),target.getNum());
				target.getLineage().descendFrom(mother);
				mother.getLineage().newMainlandSpecies(mother, distri);
				mother.getLineage().adjustTime(time);
				//System.out.println("Mainland Species: ");
				//System.out.println(mother.getSpc());
			}
			
			for (int i=0; i<N; i++){
				if (i == iPC){
					count++;
				}
				else if (i<iPC){			// Islands with number < iPC
					if (saut == i+3){			
						//System.out.println("External recruitment");
						int iPP = count;
						actPat[iPC]--;
						actPat[iPP]++;
						int positionCible = activ.indexOf(target);
						Individual mother = new Individual(iPP,target.getNum());
						activ.remove(ind_cible);
						activ.add(positionCible,mother);				// The mother is inactive at t, but was active at t-1
						target.getLineage().descendFrom(mother);
						//System.out.println("Island of origin:");
						//System.out.println(iPP);
						sample[target.getNum()].setLoc(iPP);
						event = 1;
						from = iPC;
					}
					else if (saut == i+2+N){	
						//System.out.println("External coalescence");
						int patch = count;								// Origin of the mother
						activ.remove(ind_cible);
						nbActiv--;
						actPat[iPC]--;
						Iterator<Individual> it = activ.iterator();  	// Iterator = class to got through all the element of a collection
						int compte = 0;
						r = Math.random();
						int indNb = (int)Math.floor(r*actPat[patch]);
						Individual mother = target; 					// Initial value without any importance
						boolean trouve = false;
						while(!trouve){									// To find out with which active individual in the patch the target individual will coalesce
							if (actPat[patch] != 1){
								if(it.hasNext()) {mother = it.next();}							
								else {throw(new Exception("Wrong assessment of the number of active individuals outside island "+iPC));}
								if(mother.getLoc() == patch){
									if (compte == indNb) {trouve = true;}
									else {compte++;}
								}										
								
							}
							else {
								if (it.hasNext()){
									mother = it.next();
									while (mother.getLoc() != patch) {mother = it.next();}
								}
								else {throw(new Exception("Wrong assessment of the number of active individuals outside island "+iPC));}								
								trouve = true;
							}
						}
						//System.out.println("Index of the mother:");
						//System.out.println(mother.getNum());
						event = 2;
						from = target.getNum();
						position = target.getLoc();
						motherNum = mother.getNum();
						mother.getLineage().mother(target.getLineage().getSampl());
						for (int i1=0; i1<mother.getLineage().getSampl().size();i1++){
							if (mother.getLineage().getSampl().get(i1).getNum() != target.getNum()){
								sample[target.getNum()].setTopInd(-1);
							}
						}
						
					}
					count++;
				}
				else {						// Islands with number > iPC
					if (saut==i+2){			
						//System.out.println("External recruitment");
						int iPP = count;
						actPat[iPC]--;
						actPat[iPP]++;
						Individual mother = new Individual(iPP,target.getNum());
						int positionCible = activ.indexOf(target);
						activ.remove(ind_cible);
						activ.add(positionCible,mother);				
						target.getLineage().descendFrom(mother);
						//System.out.println("Island of origin:");
						//System.out.println(iPP);
						sample[target.getNum()].setLoc(iPP);
						event = 1;
						from = iPC;
					}
					else if (saut==i+1+N){	
						//System.out.println("External coalescence");
						int patch = count;								
						activ.remove(ind_cible);
						nbActiv--;
						actPat[iPC]--;
						Iterator<Individual> it = activ.iterator();  
						int compte = 0; 
						Individual mother = target; 
						r = Math.random();
						int indNb = (int)Math.floor(r*actPat[patch]);			
						boolean trouve = false;
						while(!trouve){
							if (actPat[patch] != 1){
								if(it.hasNext()) {mother = it.next();}							
								else {throw(new Exception("Wrong assessment of the number of active individuals outside island "+iPC));}
								if(mother.getLoc() == patch){
									if (compte == indNb) {trouve = true;}
									else {compte++;}
								}	
							}
							else {
								if (it.hasNext()){
									mother = it.next();
									while (mother.getLoc() != patch) {mother = it.next();}
								}
								else {throw(new Exception("Wrong assessment of the number of active individuals outside island "+iPC));}								
								trouve = true;
							}
						}	
						event = 2;
						from = target.getNum();
						position = target.getLoc();
						motherNum = mother.getNum();
						mother.getLineage().mother(target.getLineage().getSampl());
						//System.out.println("Index of the mother:");
						//System.out.println(mother.getNum());
						for (int i1=0; i1<mother.getLineage().getSampl().size();i1++){
							if (mother.getLineage().getSampl().get(i1).getNum() != target.getNum()){
								sample[target.getNum()].setTopInd(-1);
							}
						}						
					}	
					count++;
				}
			}
		}
	}

// I/O:

	public String sumUp(){
		StringBuffer s = new StringBuffer();
		s.append("Geography: Ji = ");
		s.append(J);
		s.append(", N = ");
		s.append(N);
		s.append(", Carrying capacity of the islands:");
		for(int i=0; i<N;i++){
			s.append(" ");
			s.append(CCPat[i]);
		}
		s.append(", Number of active individuals: ");
		s.append(nbActiv);
		s.append(", Time steps (number of events): ");
		s.append(time);					// Print time "after this event (i.e. number of events)"
/*		s.append("\nActive individuals:\n");
		for(int i=0; i<nbActiv;i++){
			s.append("   ");
			s.append(activ.get(i).sumUp());
			s.append("\n");
		}*/
		s.append("\n");
		s.append("State of the sample:\n");
		s.append("   Islands: " + this.asso_isld() + "\n");		// Be careful: once an individual is absorbed into a lineage and is inactive, its island index stops changing: it stays the one it was when in underwent coalescence, even if that of its mother continues changing
		s.append("   Species: " + this.sumUp_esp() + "\n");
		s.append("   Time step of migration from the mainland (overestimated): " + this.sumUp_time() + "\n");
		s.append("   Top individuals: " + this.sumUp_topInd() + "\n\n");
		return(s.toString());
	}

	public String asso_isld(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getLoc()+",");
		}
		s.append(sample[nbInd-1].getLoc());
		return(s.toString());
	}

	public String sumUp_esp(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getSpc()+",");
		}
		s.append(sample[nbInd-1].getSpc());
		return(s.toString());
	}

	public String sumUp_time(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getTime()+",");
		}
		s.append(sample[nbInd-1].getTime());
		return(s.toString());
	}

	public String sumUp_topInd(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getTopInd()+",");
		}
		s.append(sample[nbInd-1].getTopInd());		
		return(s.toString());
	}

// TEST (example _test): 3 islands, 1 mainland; 30 individuals per island, with 10% of them sampled and followed through coalescence

	public static void main(String args[]) throws Exception{        
		
		for(int simu=0; simu<nbSim; simu++){										// Number of simulation replicates: TO BE MODIFIED
			Coalescent coa = new Coalescent();
			//System.out.println(coa.sumUp());
			while(coa.nbActifs()>0){
				ReadTable.writeOutput(sample,target,motherNum,time,event,from,position, nbActiv, simu);
				coa.cycle();
				//System.out.println(coa.sumUp());
			}
			ReadTable.writeLastOutput(sample,target,time,event,from, simu);
		}	
	}
}
