package outils;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

// This algorithm runs the coalescence under a model of unstructured archipelago: it builds (backward in time) 
// the genealogy of a sample of the individuals living within the archipelago at present time, under neutrality hypotheses, 
// and following the migration of their ancestral lineages among islands back in time, until reaching the immigration events from the mainland.
// Unlike "coa_struct", you may here consider multiple independent unstructured archipelagos at a time (each row in "Mconnectivity_test.txt" = 1 archipelago), 
// generating one output file "coaHist_unstruct_test_ARCHIP.txt" for each archipelago ARCHIP. Here, this file contains all simulation replicates. This decreases the number of simulations to run.
// Fix the number of simulation replicates, nbSim, as you want.
// The word "patch(es)" used below refers to archipelago(s).

// Input files and output files are read/produces using "ReadTable.java"; explanations about inputs and outputs are given in this java file.
// In this study example, we consider simplified input files (few islands, few individuals, simple theoretical matrix of connectivity between landmasses), 
// named "name_test.txt". Once you understand to code, make your own input files.
// We left (inactivated, commented) commands System.out.println(): they enable to print events on screen, and you may want to re-activate them to better understand how the code works.

public class Coalescent {

// ATTRIBUTES:

	int J;                     			// Number of individuals in the archipelago 
	static int[] CCPat;                	// Carrying capacity of the patches
	static double[] freq;				// Sampling frequencies in each patch
	static double[][] Mcon;				// Matrix of connectivity with the mainland, for all archipelagos
	static int nbArchip;					
	static int[] distri;				// Mainland species abundance distribution
	static int time;					// time step						
	static Individual[] sample;			// Spatialized sample of individuals followed by the coalescence
	static Individual target;			// Individual randomly chosen at each time step, target individual

	static int event;					// Informations to write the output file
	static int from;					
	static int motherNum;
	static int position;

	LinkedList<Individual> activ;       // List of the individuals in the coalescent ("active" individuals; list updated at each time step; its size decreases as time goes backward)
	static int nbActiv;                 // Number of individuals in this list 
	int[] actPat;                      	// Number of individuals in the coalescent which are located in patch i

	static double[] transition;			// Vector of transition probabilities, which allows simulating the markovian dynamic of the coalescent. 
										// Knowing which individual is just born, these probabilities are
											// 0 -> probability of local coalescence: the mother of the born individual is another traced individual, from the same patch
											// 1 -> probability of local recruitment: the mother of the born individual is an individual not traced (not sampled), from the same patch
											// 2 -> probability of migration from the mainland: the mother of the born individual is an individual from the mainland

	static int nbSim = 10;						// Number of simulation replicates: TO BE MODIFIED
	
// CONSTRUCTOR:

	public Coalescent(int archip) throws Exception{
				
															
		int nbSampl = (int)Math.floor(CCPat[archip]*freq[archip]);		// sampling individuals
		J = CCPat[archip];
		if(nbSampl>CCPat[archip]) throw(new Exception("Over sampling of archipelago "+archip));			
		
		nbActiv = nbSampl;
		actPat = new int[1];					// kept as a vector to conserve a similar structure as in "coa_struct"
		actPat[0]=nbSampl;
		
		activ = new LinkedList<Individual>();
		Individual[] samp = new Individual[nbSampl];
		int curseur = 0;
		for(int j = 0 ; j<nbSampl ; j++){
			Individual ind = new Individual(0,curseur);
			Lineage lign = new Lineage(ind);
			activ.add(ind);
			samp[curseur]=ind;
			curseur++;
		}
		sample = samp;		
		target = sample[0];
		motherNum = 0;
		event = 0;
		from = 0;
		position = -1;
		time = 0;
		transition = new double[3];
	}

// READERS:

	public int nbActifs(){
		return(nbActiv);
	}

// METHODS:

	public void cycle(int archip) throws Exception{
		
		time++;
		position=-1;
		if (nbActiv<=0) throw(new Exception("Coalescence already finished"));
		
		else if(nbActiv==1){			// If the coalescent contains only one individual, we look for the mainland ancestor to this individual.
			target = activ.get(0);			// the target individual is the last individual within the list of active individuals
			activ.remove(0);				// list "activ" updated
			nbActiv--;					
			int iPC = target.getLoc();		// iPC = index of the patch of the target individual
			actPat[iPC]--;              	// list "actPat" updated
			event = 3;						// Migration from the mainland
			from = target.getNum();
			
			Individual mother = new Individual(target.getLoc(),target.getNum());	// Choosing the mainland ancestor
			target.getLineage().descendFrom(mother);
			mother.getLineage().newMainlandSpecies(mother, distri);
			
			mother.getLineage().adjustTime(time); 
			//System.out.println("Mainland species: ");
			//System.out.println(mother.getSpc());
		}

		else{

			// STEP 1 : Choice of the target individual

			double r = Math.random();
			int ind_cible = (int)Math.floor(r*(double)nbActiv); 	// Integer within [0,nbActiv-1]: index of the target individual
			target = activ.get(ind_cible);

			// STEP 2 : Calculation of the transition vector, knowing which individual is just born
			
			int iPC = target.getLoc();								// iPC = index of the patch of the target individual
			int KPC = CCPat[iPC];									// Carrying capacity of the patch of the target individual
			
			double m = Mcon[archip][0];								// Probability of coming from the same patch		
			transition[0]=m*(actPat[iPC]-1)/((double) (KPC-1));		// Local coalescence
			transition[1]=m*(KPC-actPat[iPC])/((double) (KPC-1));	// Local recruitment
			transition[2]=Mcon[archip][1];									// Migration from the mainland

			//System.out.println("Index of the target individual:");
			//System.out.println(target.getNum());
			
			// STEP 3 : Choice of the transition
			
			r = Math.random();
			int saut = 0;
			double somme_cum = transition[0];
			while(somme_cum<r){
				saut++;
				somme_cum = somme_cum+transition[saut]; 
			}
			
			// STEP 4 : Update of the system according to the transition
			
			if(saut == 0){
				//System.out.println("Local coalescence");
				r = Math.random();
				int iDPP = (int)Math.floor(r*(actPat[iPC]-1));  // Index of the active mother, within the non-target active individuals of the patch iPC (integer within [0,actPat[iPC]-1])
				activ.remove(ind_cible);
				nbActiv--;
				actPat[iPC]--;
				Iterator<Individual> it = activ.iterator();
				int compte = -1;
				Individual mother = target; 					// Initial value without any importance
				boolean trouve = false;
				while(!trouve){
					if(!it.hasNext()) throw(new Exception("Wrong assessment of the number of active individuals inside patch "+iPC));
					else mother = it.next();
					if(mother.getLoc() == iPC){
						compte++;
					}
					trouve = (compte == iDPP);
				}
				//System.out.println("ancestor.getNum:");
				//System.out.println(mother.getNum());
				event = 2;
				motherNum = mother.getNum();
				from = target.getNum();
				position = target.getLoc();
				mother.getLineage().mother(target.getLineage().getSampl());
				for (int i1 = 0; i1<mother.getLineage().getSampl().size();i1++){
					if (mother.getLineage().getSampl().get(i1).getNum() != target.getNum()){
						sample[target.getNum()].setTopInd(-1);
					}
				}
				
			}
			else if(saut == 1){				
				event = 0;
				from = 0;
				//System.out.println("Local recruitment");		// The mother has the same characteristics as the target individual, so nothing changes in the macro data of the coalescent (the 2 following lines are thus useless)
				Individual mother = new Individual(target.getLoc(),target.getNum());
				target.getLineage().descendFrom(mother);				
			}
			
			else if (saut == 2){
				event = 3;
				from = target.getNum();
				//System.out.println("Migration from the mainland");	
				actPat[iPC]--;
				activ.remove(ind_cible);
				nbActiv--;	
				Individual mother = new Individual(target.getLoc(),target.getNum());
				target.getLineage().descendFrom(mother);
				mother.getLineage().newMainlandSpecies(mother, distri);
				mother.getLineage().adjustTime(time);
				//System.out.println("Mainland Species: ");
				//System.out.println(mother.getSpc());
				
			}		
		}
	}

// I/O:

	public String sumUp(){
		StringBuffer s = new StringBuffer();
		s.append("Geography: Ji = ");
		s.append(J);
		s.append(", N = ");
		s.append(1);
		s.append(", Carrying capacity of the patch:");
		s.append(CCPat);
		s.append(", Number of active individuals: ");
		s.append(nbActiv);
		s.append(", Time steps (number of events): ");
		s.append(time);					// On affiche le temps "aprs cet vnement" (i.e = le nombre d'vnements)
/*		s.append("\nActive individuals:\n");
		for(int i=0; i<nbActiv;i++){
			s.append("   ");
			s.append(activ.get(i).sumUp());
			s.append("\n");
		}*/
		s.append("\n");
		s.append("State of the sample:\n");
		s.append("   Patch: " + this.asso_patch() + "\n");		// Be careful: once an individual is absorbed into a lineage and is inactive, its patch number stops changing: it stays the one it was when in underwent coalescence, even if that of its mother continues changing
		s.append("   Species: " + this.sumUp_esp() + "\n");
		s.append("   Time step of migration from the mainland (overestimated): " + this.sumUp_time() + "\n");
		s.append("   Top individuals: " + this.sumUp_topInd() + "\n\n");
		return(s.toString());
	}

	public String asso_patch(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getLoc()+",");
		}
		s.append(sample[nbInd-1].getLoc());
		return(s.toString());
	}

	public String sumUp_esp(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getSpc()+",");
		}
		s.append(sample[nbInd-1].getSpc());
		return(s.toString());
	}

	public String sumUp_time(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getTime()+",");
		}
		s.append(sample[nbInd-1].getTime());
		return(s.toString());
	}

	public String sumUp_topInd(){
		StringBuffer s = new StringBuffer();
		s.append("+");
		int nbInd= sample.length;
		for(int i=0;i<nbInd-1;i++){
			s.append(sample[i].getTopInd()+",");
		}
		s.append(sample[nbInd-1].getTopInd());		
		return(s.toString());
	}

// TEST (example): 3 insular patches, 1 mainland; 4 individuals per patch (half of them sampled and followed through coalescence)

	public static void main(String args[]) throws Exception{  
		
		Mcon = ReadTable.readConnectivity();
		distri = ReadTable.readDistri();		
		CCPat = ReadTable.readK();
		freq = ReadTable.readSamplingFreq();
		int nbArchip=CCPat.length;									// Number of archipelagos considered at the same time
		
		for (int archip = 0; archip < nbArchip; archip++) {
		  for (int simu=0; simu<nbSim; simu++){	
			Coalescent coa = new Coalescent(archip);
			//System.out.println(coa.sumUp());
			while(coa.nbActifs()>0){
				ReadTable.writeOutput(sample,target,motherNum,time,event,from,position, nbActiv, archip);
				coa.cycle(archip);
				//System.out.println(coa.sumUp());
			}
			ReadTable.writeLastOutput(sample,target,time,event,from, archip);
		  }
		}
	}
}
