package outils;

import java.util.ArrayList;
import java.util.StringTokenizer;

// This file takes, for each simulation replicate SIM, the history of the coalescence of sampled individuals in a structured archipelago 
// as an input ("coaHist_struct_test_SIM.txt"), and works it forward in time to assign species to present individuals, 
// either without speciation (by simple succession from their mainland ancestor), storing results in "NoSpc_struct_test.txt",
// or with protracted speciation (determining where and when speciation events took place), storing results in "SpcProtr_struct_test.txt". 
// Fix the expected time to protracted speciation in the archipelago, Tspc_archip. Its value on each island is proportional to the number of individuals.
// Fix the number of simulation replicates, nbSim (similar to the one used in "coa_struct").

// Input files and output files are read/produces using "ReadWrite.java"; explanations about inputs and outputs are given in this java file.
// We left (inactivated, commented) commands System.out.println(): they enable to print events on screen, and you may want to re-activate them to better understand how the code works.

public class Speciation{	
	
	public static void main(String args[]) throws Exception{
			
	  // STEP 1: READ & INITIALIZING DATA COMMON TO ALL SIMULATION REPLICATES
		
	  double[] K = ReadWrite.readK();						// carrying capacity on each island
	  double J = 0;											// carrying capacity on the archipelago
	  for (int i=0; i<K.length; i++){						
		J = J + K[i];
		//System.out.println(K[i]);
	  }
	  //System.out.println(J);
	  
	  double[][] Mcon = ReadWrite.readConnectivity(K.length+1);	// matrix of connectivity between islands
	  int[] mainlandSp = ReadWrite.readDistri();			// cumulative species abundance distribution on the mainland
	  double Tspc_archip = 1000000;						// time needed to complete speciation on the archipelago: TO BE MODIFIED
	  double[] Tspc_islds = K;								// time needed to complete speciation on each island
	  for (int i = 0; i < Tspc_islds.length; i++) {
		  Tspc_islds[i] = Tspc_archip*Tspc_islds[i]/J;
		  //Tspc_islds[i] = Tspc_archip;					//TEST
		  //System.out.println(Tspc_islds[i]);
	  }
	  int nbSim = 10;										// number of simulation replicates: TO BE MODIFIED
			  
	  //  2: RUN FOR EACH SIMULATION REPLICATE
	  
	  for (int simu = 0; simu < nbSim; simu++) {			//int simu = nbSim; simu <= nbSim+10; simu++
		
		    // STEP 2.A: PREPARING THE DATA
		  
		    ArrayList<Integer> sameTime = new ArrayList<Integer>();
		    ArrayList<Integer> firstState = new ArrayList<Integer>();
		    ArrayList<Integer> lastState = new ArrayList<Integer>();
		    ArrayList<Integer> cladeTemp = new ArrayList<Integer>();
			
			double[][] input = ReadWrite.readInput(simu);					// input = input "coaHist_...SIM.txt", for this simulation replicate SIM 
			String[] clade = ReadWrite.readClade(input.length, simu);		// clade = clade in "coaHist_...SIM.txt"

			firstState.clear();
			lastState.clear();				
			int stepMax = (int) input[input.length-1][0];						// Number of time steps in this simulation
			
			int Na = 0;															// Na = number of active individuals (dynamic)
			for (int i=1; i<=stepMax;i++){										// Fill up the 4th column: "Time"
				for (int j=0;j<(input.length-1);j++){
					if (input[j][0] == i){
						Na = (int) input[j][9];
						sameTime.add(j);										// If several lines in input with the same "time step"
					}
				}						
				double rexp = StdRandom.exp((double)Na/J);						// Exponential law (approximation, especially when high proportion of traced individuals (high frequency of sampling and beginning of the coalescence))
				for (int i1=0;i1<sameTime.size();i1++){
					input[sameTime.get(i1)][3] = rexp;
				}
				sameTime.clear();
			}			
			/*double su = 0;													// To know the total time of coalescence
			for (int i1=0; i1<input.length;i1++){
				su = su+input[i1][3];
			}
			System.out.println(su);*/
				
			for (int i=0; i<input.length;i++){									// firstState = lines of the individuals at t=0 (present time)
				if (input[i][0] == 0){
					firstState.add(i);
				}
			}
			
			for (int i=0; i<firstState.size(); i++){							// Fill up 6th column: for all the individuals of the initial state, clade in "Clade_Num" (tranform string into double)
				StringTokenizer st = new StringTokenizer(clade[firstState.get(i)],"-");
				String sp = st.nextToken();
				int sp2 = Integer.parseInt(sp);
				cladeTemp.add(sp2);												
				input[firstState.get(i)][5] = cladeTemp.get(0); 
				cladeTemp.clear();
			}
		
			// STEP 2.B: ATTRIBUTING SPECIES WITH NO SPECIATION		
			
			for (int i=0; i<input.length;i++){									// lastState = lines which have time step = last time step
				if (input[i][0] == stepMax){
					lastState.add(i);
				}
			}
			
			for (int i=0;i<lastState.size();i++){								// for all these lines, find the species of the mainland ancestor, and attribute it to all the corresponding individuals
				StringTokenizer st = new StringTokenizer(clade[lastState.get(i)],"-");
				String sp = st.nextToken();
				int sp2 = Integer.parseInt(sp);
				cladeTemp.add(sp2);
				while (st.hasMoreTokens()){			
					sp = st.nextToken();
					sp2 = Integer.parseInt(sp);
					cladeTemp.add(sp2);
				}
				for (int j=0; j<cladeTemp.size();j++){
					for (int k=0; k<firstState.size();k++){
						if (input[firstState.get(k)][5] == cladeTemp.get(j)){
							input[firstState.get(k)][4] = input[lastState.get(i)][2];		// Species without speciation
						}
					}
				}
				cladeTemp.clear();
			}
			
			ReadWrite.writeNoSpeciation(input,firstState, simu);
			
			// STEP 2.C: ATTRIBUTING SPECIES WITH PROTRACTED SPECIATION
			
			// INITIALIZING DATA
			int nbSpec = 0;  									// Counter of the number of speciation events (to know what species number to give when there is a speciation event)
			double species = 0;									// A species index, useful for the algorithm
			int nbEspApprox = mainlandSp.length;				// Number of species on the mainland	 
			double connectivity = 0;							// Connectivity of an island (percentage of external propagules) 
			double geneFlow = 0;								// Time delay of speciation due to gene flow
			int first_col = 0; 									// 0 if individuals of the same species were already present on the island, 1 otherwise (first colonization)
			
			int spMax = (int) input[input.length-1][2];			
			for (int i=0; i<lastState.size();i++){
				if (input[lastState.get(i)][2]>spMax){
					spMax = (int) input[lastState.get(i)][2];	// spMax = maximal species index on the mainland
				}
			}
											 
			double[] Species = new double[firstState.size()];	// Vector of the species number of all individuals; updated while going through the history of coalescence
			for (int i=0;i<Species.length;i++){
				Species[i] = -1;
			}	
			double[] Position = new double[firstState.size()];	// Vector of the position of all individuals; updated while going through the history of coalescence
			for (int i=0;i<Position.length;i++){
				Position[i] = -1;
			}
			double[] Event = new double[firstState.size()];		// Vector of the event of all individuals (if the individual has already been updated this time step or not); updated while going through the history of coalescence
			for (int i=0;i<Event.length;i++){
				Event[i] = 0;
			}		
			double[] Clock = new double[firstState.size()];	    // Clock of all individuals: starts at Tspc when a new species arrives on an island, and decreases as time goes forward; speciation when clock <= 0
			for (int i=0;i<Clock.length;i++){
				Clock[i] = -999;
			}
			
			// WORKING FORWARD IN TIME THE HISTORY OF COALESCENCE, UPDATING THE VECTORS OF CHARACTERISTICS OF THE INDIVIDUALS
			for (int i=input.length-1; i>=0;i--){		

				// STEP 2.C.1: If new time step (and not last one), do all necessary speciation events...
				
				if (i < (input.length-lastState.size())){
					if (input[i][0] != input[i+1][0]) {			
						for (int j=0;j<Event.length;j++){		// Initialisation of Event at 0
							Event[j] = 0;
						}
						for (int j=0;j<Species.length;j++){   	// All active clocks (of individuals active at this time step), are updated = decreased according to the time since the last event (Time)
							if (Clock[j] != -999){Clock[j] -= input[i+1][3];}
						}
						
						for (int j=0;j<Species.length;j++){ 								
							if (Species[j] != -1 & Event[j] == 0 & Clock[j] < 0 & Clock[j] != -999){		// Individual active, species not yet attributed and clocks < 0)
								//System.out.println("Speciation");
								species = Species[j];
						        connectivity = 1 - Mcon[(int) input[i][1]][(int) (input[i][1])];				// Islands are modelled all together
								geneFlow = StdRandom.poisson(Tspc_islds[(int) input[i][1]]*connectivity);					// Delay of speciation due to the gene flow, depends on the connectivity of the island 
								for (int k=0; k<Species.length;k++){
									if (Species[k] == species & Position[k] == Position[j]){
										Event[k] = 1;
										Species[k] = (double) (nbEspApprox+1+nbSpec);
										Clock[k] = Tspc_islds[(int) input[i][1]] + geneFlow;
									}
								}
								nbSpec++;							
							}	
						}
					}
				}
				
				// STEP 2.C.2: If Event is not a local recruitment (in that case, only the clocks decrease and speciation events may occur): update all the vectors of characteristics of the individuals 
				
				if (input[i][6] != 0){   								
					
					StringTokenizer st = new StringTokenizer(clade[i],"-");
					String sp = st.nextToken();
					int sp2 = Integer.parseInt(sp);
					cladeTemp.add(sp2);																		// Find out the descendants of this lineage
					while (st.hasMoreTokens()){			
						sp = st.nextToken();
						sp2 = Integer.parseInt(sp);
						cladeTemp.add(sp2);
					}
					
					if (input[i][6] == 3) {		// Migration from the mainland
						
						for (int k=0; k<Species.length;k++){
							if (Species[k] == Species[cladeTemp.get(0)] & Position[k] == input[i][1]){			// If this species was already present on the island: the new individual takes the clocks of the individuals of the same species already present on the island 
								for (int j=0;j<cladeTemp.size();j++){
									Clock[cladeTemp.get(j)] = Clock[k];
									Species[cladeTemp.get(j)] = input[i][2];
									Position[cladeTemp.get(j)] = input[i][1];
								}
							}				
						}
						if (Clock[cladeTemp.get(0)] == -999) {                                					// If this species was not present on the island: the new individuals starts differentiating
							connectivity = 1 - Mcon[(int) input[i][1]][(int) (input[i][1])];						// Islands are modelled all together
							geneFlow = StdRandom.poisson(Tspc_islds[(int) input[i][1]]*connectivity);	// Time delay of speciation due to gene flow (for this island)
							for (int j=0;j<cladeTemp.size();j++){
								Species[cladeTemp.get(j)] = input[i][2];
								Clock[cladeTemp.get(j)] = Tspc_islds[(int) input[i][1]]+geneFlow;
								Position[cladeTemp.get(j)] = input[i][1];
							}
						}
						cladeTemp.clear();
					}
					
					else if (input[i][6] == 2){		// Coalescence	
						int j = 0;					
						if (input[i][1] != input[i][7]){	// External coalescence	
							while(cladeTemp.get(j) != input[i][7]){  											// Mother clade: same position
								j++;
							}
							first_col = 1;
							for (int k=0; k<Species.length;k++){
								if (Species[k] == Species[cladeTemp.get(j)] & Position[k] == input[i][8]){		// If this species was already present on the island; if not, the new individuals keeps its clock
									first_col = 0; 
									for (int f=j;f<cladeTemp.size();f++){
										Clock[cladeTemp.get(f)] = Clock[k];	
									}
									
								}
							}
							// ADDED
							if (first_col == 1) {
								connectivity = 1 - Mcon[(int) input[i][1]][(int) (input[i][1])];						// Islands are modelled all together
								geneFlow = StdRandom.poisson(Tspc_islds[(int) input[i][1]]*connectivity);	// Time delay of speciation due to gene flow (for this island)	
								for (int f=j;f<cladeTemp.size();f++){
									Clock[cladeTemp.get(f)] = Tspc_islds[(int) input[i][1]]+geneFlow;	
								}
							}
							// END_ADDED
							for (int f=j;f<cladeTemp.size();f++){  															// Daughter clade: different position
								Position[cladeTemp.get(f)] = input[i][8];
							}						
						}																						
						cladeTemp.clear();
					}																							// If local coalescence, the position and clock are not modified
					
					else if (input[i][6] == 1){		// External recruitment	
						first_col = 1;
						for (int k=0; k<Species.length;k++){
							if (Species[k] == Species[cladeTemp.get(0)] & Position[k] == input[i][7]){			// If this species was already present on the island; if not, the new individuals keeps its clock
								first_col = 0; 
								for (int j=0;j<cladeTemp.size();j++){
									Clock[cladeTemp.get(j)] = Clock[k];	
								}
							}
						}
						// ADDED
						if (first_col == 1) {
							connectivity = 1 - Mcon[(int) input[i][1]][(int) (input[i][1])];						// Islands are modelled all together
							geneFlow = StdRandom.poisson(Tspc_islds[(int) input[i][1]]*connectivity);	// Time delay of speciation due to gene flow (for this island)	
							for (int j=0;j<cladeTemp.size();j++){
								Clock[cladeTemp.get(j)] = Tspc_islds[(int) input[i][1]]+geneFlow;	
							}
						}
						// END_ADDED
						for (int j=0; j<cladeTemp.size();j++){
							Position[cladeTemp.get(j)] = input[i][7];
						}
						cladeTemp.clear();
					}
				}
			}
			ReadWrite.writeSpcProtracted(Position,Species, simu); 
					
	    }
    }
}

